/*
 * Copyright (C) 2002 Roman Zippel <zippel@linux-m68k.org>
 * Released under the terms of the GNU GPL v2.0.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#define LKC_DIRECT_LINK
#include "lkc.h"

struct expr *expr_alloc_symbol(struct symbol *sym)
{
	struct expr *e = malloc(sizeof(*e));
	memset(e, 0, sizeof(*e));
	e->token = T_WORD;
	e->left.sym = sym;
	return e;
}

struct expr *expr_alloc_one(int token, struct expr *ce)
{
	struct expr *e = malloc(sizeof(*e));
	memset(e, 0, sizeof(*e));
	e->token = token;
	e->left.expr = ce;
	return e;
}

struct expr *expr_alloc_two(int token, struct expr *e1, struct expr *e2)
{
	struct expr *e = malloc(sizeof(*e));
	memset(e, 0, sizeof(*e));
	e->token = token;
	e->left.expr = e1;
	e->right.expr = e2;
	return e;
}

struct expr *expr_alloc_comp(int token, struct symbol *s1, struct symbol *s2)
{
	struct expr *e = malloc(sizeof(*e));
	memset(e, 0, sizeof(*e));
	e->token = token;
	e->left.sym = s1;
	e->right.sym = s2;
	return e;
}

struct expr *expr_alloc_and(struct expr *e1, struct expr *e2)
{
	if (!e1)
		return e2;
	return e2 ? expr_alloc_two('&', e1, e2) : e1;
}

struct expr *expr_copy(struct expr *org)
{
	struct expr *e;

	if (!org)
		return NULL;

	e = malloc(sizeof(*org));
	memcpy(e, org, sizeof(*org));
	switch (org->token) {
	case T_WORD:
		e->left = org->left;
		break;
	case '!':
		e->left.expr = expr_copy(org->left.expr);
		break;
	case '=':
	case T_UNEQUAL:
		e->left.sym = org->left.sym;
		e->right.sym = org->right.sym;
		break;
	case '&':
	case '|':
	case '^':
		e->left.expr = expr_copy(org->left.expr);
		e->right.expr = expr_copy(org->right.expr);
		break;
	default:
		printf("can't copy token %d\n", e->token);
		free(e);
		e = NULL;
		break;
	}

	return e;
}

void expr_free(struct expr *e)
{
	if (!e)
		return;

	switch (e->token) {
	case T_WORD:
		break;
	case '!':
		expr_free(e->left.expr);
		return;
	case '=':
	case T_UNEQUAL:
		break;
	case '|':
	case '&':
		expr_free(e->left.expr);
		expr_free(e->right.expr);
		break;
	default:
		printf("how to free token %d?\n", e->token);
		break;
	}
	free(e);
}

static int trans_count;

#define e1 (*ep1)
#define e2 (*ep2)

static void __expr_eliminate_eq(int token, struct expr **ep1, struct expr **ep2)
{
	if (e1->token == token) {
		__expr_eliminate_eq(token, &e1->left.expr, &e2);
		__expr_eliminate_eq(token, &e1->right.expr, &e2);
		return;
	}
	if (e2->token == token) {
		__expr_eliminate_eq(token, &e1, &e2->left.expr);
		__expr_eliminate_eq(token, &e1, &e2->right.expr);
		return;
	}
	if (e1->token == T_WORD && e2->token == T_WORD &&
	    e1->left.sym == e2->left.sym && (e1->left.sym->flags & (SYMBOL_YES|SYMBOL_NO)))
		return;
	if (!expr_eq(e1, e2))
		return;
	trans_count++;
	expr_free(e1); expr_free(e2);
	switch (token) {
	case '|':
		e1 = expr_alloc_symbol(&symbol_no);
		e2 = expr_alloc_symbol(&symbol_no);
		break;
	case '&':
		e1 = expr_alloc_symbol(&symbol_yes);
		e2 = expr_alloc_symbol(&symbol_yes);
		break;
	}
}

void expr_eliminate_eq(struct expr **ep1, struct expr **ep2)
{
	if (!e1 || !e2 || e1->token != e2->token)
		return;
	__expr_eliminate_eq(e1->token, ep1, ep2);
	e1 = expr_eliminate_yn(e1);
	e2 = expr_eliminate_yn(e2);
}

#undef e1
#undef e2

int expr_eq(struct expr *e1, struct expr *e2)
{
	int res, old_count;

	if (e1->token != e2->token)
		return 0;
	switch (e1->token) {
	case '=':
	case T_UNEQUAL:
		return e1->left.sym == e2->left.sym && e1->right.sym == e2->right.sym;
	case T_WORD:
		return e1->left.sym == e2->left.sym;
	case '!':
		return expr_eq(e1->left.expr, e2->left.expr);
	case '&':
	case '|':
		e1 = expr_copy(e1);
		e2 = expr_copy(e2);
		old_count = trans_count;
		expr_eliminate_eq(&e1, &e2);
		res = (e1->token == T_WORD && e2->token == T_WORD &&
		       e1->left.sym == e2->left.sym);
		expr_free(e1);
		expr_free(e2);
		trans_count = old_count;
		return res;
	}

	print_expr(0, e1, 0);
	printf(" = ");
	print_expr(0, e2, 0);
	printf(" ?\n");

	return 0;
}

struct expr *expr_eliminate_yn(struct expr *e)
{
	struct expr *tmp;

	if (e) switch (e->token) {
	case '&':
		e->left.expr = expr_eliminate_yn(e->left.expr);
		e->right.expr = expr_eliminate_yn(e->right.expr);
		if (e->left.expr->token == T_WORD) {
			if (e->left.expr->left.sym == &symbol_no) {
				expr_free(e->left.expr);
				expr_free(e->right.expr);
				e->token = T_WORD;
				e->left.sym = &symbol_no;
				e->right.expr = NULL;
				return e;
			} else if (e->left.expr->left.sym == &symbol_yes) {
				free(e->left.expr);
				tmp = e->right.expr;
				*e = *(e->right.expr);
				free(tmp);
				return e;
			}
		}
		if (e->right.expr->token == T_WORD) {
			if (e->right.expr->left.sym == &symbol_no) {
				expr_free(e->left.expr);
				expr_free(e->right.expr);
				e->token = T_WORD;
				e->left.sym = &symbol_no;
				e->right.expr = NULL;
				return e;
			} else if (e->right.expr->left.sym == &symbol_yes) {
				free(e->right.expr);
				tmp = e->left.expr;
				*e = *(e->left.expr);
				free(tmp);
				return e;
			}
		}
		break;
	case '|':
		e->left.expr = expr_eliminate_yn(e->left.expr);
		e->right.expr = expr_eliminate_yn(e->right.expr);
		if (e->left.expr->token == T_WORD) {
			if (e->left.expr->left.sym == &symbol_no) {
				free(e->left.expr);
				tmp = e->right.expr;
				*e = *(e->right.expr);
				free(tmp);
				return e;
			} else if (e->left.expr->left.sym == &symbol_yes) {
				expr_free(e->left.expr);
				expr_free(e->right.expr);
				e->token = T_WORD;
				e->left.sym = &symbol_yes;
				e->right.expr = NULL;
				return e;
			}
		}
		if (e->right.expr->token == T_WORD) {
			if (e->right.expr->left.sym == &symbol_no) {
				free(e->right.expr);
				tmp = e->left.expr;
				*e = *(e->left.expr);
				free(tmp);
				return e;
			} else if (e->right.expr->left.sym == &symbol_yes) {
				expr_free(e->left.expr);
				expr_free(e->right.expr);
				e->token = T_WORD;
				e->left.sym = &symbol_yes;
				e->right.expr = NULL;
				return e;
			}
		}
		break;
	}
	return e;
}

/*
 * bool FOO!=n => FOO
 */
struct expr *expr_trans_bool(struct expr *e)
{
	if (!e)
		return NULL;
	switch (e->token) {
	case '&':
	case '|':
	case '!':
		e->left.expr = expr_trans_bool(e->left.expr);
		e->right.expr = expr_trans_bool(e->right.expr);
		break;
	case T_UNEQUAL:
		// FOO!=n -> FOO
		if (e->left.sym->type == SYMBOL_TRISTATE) {
			if (e->right.sym == &symbol_no) {
				e->token = T_WORD;
				e->right.sym = NULL;
			}
		}
		break;
	}
	return e;
}

/*
 * e1 || e2 -> ?
 */
struct expr *expr_join_or(struct expr *e1, struct expr *e2)
{
	struct expr *tmp;
	struct symbol *sym1, *sym2;

	if (expr_eq(e1, e2))
		return expr_copy(e1);
	if (e1->token != '=' && e1->token != T_UNEQUAL && e1->token != T_WORD && e1->token != '!')
		return NULL;
	if (e2->token != '=' && e2->token != T_UNEQUAL && e2->token != T_WORD && e2->token != '!')
		return NULL;
	if (e1->token == '!') {
		tmp = e1->left.expr;
		if (tmp->token != '=' && tmp->token != T_UNEQUAL && tmp->token != T_WORD)
			return NULL;
		sym1 = tmp->left.sym;
	} else
		sym1 = e1->left.sym;
	if (e2->token == '!') {
		if (e2->left.expr->token != T_WORD)
			return NULL;
		sym2 = e2->left.expr->left.sym;
	} else
		sym2 = e2->left.sym;
	if (sym1 != sym2)
		return NULL;
	if (sym1->type != SYMBOL_BOOLEAN && sym1->type != SYMBOL_TRISTATE)
		return NULL;
	if (sym1->type == SYMBOL_TRISTATE) {
		if (e1->token == '=' && e2->token == '=' &&
		    ((e1->right.sym == &symbol_yes && e2->right.sym == &symbol_mod) ||
		     (e1->right.sym == &symbol_mod && e2->right.sym == &symbol_yes))) {
			// (a='y') || (a='m') -> (a!='n')
			return expr_alloc_comp(T_UNEQUAL, sym1, &symbol_no);
		}
		if (e1->token == '=' && e2->token == '=' &&
		    ((e1->right.sym == &symbol_yes && e2->right.sym == &symbol_no) ||
		     (e1->right.sym == &symbol_no && e2->right.sym == &symbol_yes))) {
			// (a='y') || (a='n') -> (a!='m')
			return expr_alloc_comp(T_UNEQUAL, sym1, &symbol_mod);
		}
		if (e1->token == '=' && e2->token == '=' &&
		    ((e1->right.sym == &symbol_mod && e2->right.sym == &symbol_no) ||
		     (e1->right.sym == &symbol_no && e2->right.sym == &symbol_mod))) {
			// (a='m') || (a='n') -> (a!='y')
			return expr_alloc_comp(T_UNEQUAL, sym1, &symbol_yes);
		}
	}
	if (sym1->type == SYMBOL_BOOLEAN && sym1 == sym2) {
		if ((e1->token == '!' && e1->left.expr->token == T_WORD && e2->token == T_WORD) ||
		    (e2->token == '!' && e2->left.expr->token == T_WORD && e1->token == T_WORD))
			return expr_alloc_symbol(&symbol_yes);
	}

	printf("optimize ");
	print_expr(0, e1, 0);
	printf(" || ");
	print_expr(0, e2, 0);
	printf(" ?\n");
	return NULL;
}

struct expr *expr_join_and(struct expr *e1, struct expr *e2)
{
	struct expr *tmp;
	struct symbol *sym1, *sym2;

	if (expr_eq(e1, e2))
		return expr_copy(e1);
	if (e1->token != '=' && e1->token != T_UNEQUAL && e1->token != T_WORD && e1->token != '!')
		return NULL;
	if (e2->token != '=' && e2->token != T_UNEQUAL && e2->token != T_WORD && e2->token != '!')
		return NULL;
	if (e1->token == '!') {
		tmp = e1->left.expr;
		if (tmp->token != '=' && tmp->token != T_UNEQUAL && tmp->token != T_WORD)
			return NULL;
		sym1 = tmp->left.sym;
	} else
		sym1 = e1->left.sym;
	if (e2->token == '!') {
		if (e2->left.expr->token != T_WORD)
			return NULL;
		sym2 = e2->left.expr->left.sym;
	} else
		sym2 = e2->left.sym;
	if (sym1 != sym2)
		return NULL;
	if (sym1->type != SYMBOL_BOOLEAN && sym1->type != SYMBOL_TRISTATE)
		return NULL;

	if ((e1->token == T_WORD && e2->token == '=' && e2->right.sym == &symbol_yes) ||
	    (e2->token == T_WORD && e1->token == '=' && e1->right.sym == &symbol_yes))
		// (a) && (a='y') -> (a='y')
		return expr_alloc_comp('=', sym1, &symbol_yes);

	if ((e1->token == T_WORD && e2->token == T_UNEQUAL && e2->right.sym == &symbol_no) ||
	    (e2->token == T_WORD && e1->token == T_UNEQUAL && e1->right.sym == &symbol_no))
		// (a) && (a!='n') -> (a)
		return expr_alloc_symbol(sym1);

	if (sym1->type == SYMBOL_TRISTATE) {
		if (e1->token == '=' && e2->token == T_UNEQUAL) {
			// (a='b') && (a!='c') -> 'b'='c' ? 'n' : a='b'
			sym2 = e1->right.sym;
			if ((e2->right.sym->flags & SYMBOL_CONST) && (sym2->flags & SYMBOL_CONST))
				return sym2 != e2->right.sym ? expr_alloc_comp('=', sym1, sym2)
							     : expr_alloc_symbol(&symbol_no);
		}
		if (e1->token == T_UNEQUAL && e2->token == '=') {
			// (a='b') && (a!='c') -> 'b'='c' ? 'n' : a='b'
			sym2 = e2->right.sym;
			if ((e1->right.sym->flags & SYMBOL_CONST) && (sym2->flags & SYMBOL_CONST))
				return sym2 != e1->right.sym ? expr_alloc_comp('=', sym1, sym2)
							     : expr_alloc_symbol(&symbol_no);
		}
		if (e1->token == T_UNEQUAL && e2->token == T_UNEQUAL &&
			   ((e1->right.sym == &symbol_yes && e2->right.sym == &symbol_no) ||
			    (e1->right.sym == &symbol_no && e2->right.sym == &symbol_yes)))
			// (a!='y') && (a!='n') -> (a='m')
			return expr_alloc_comp('=', sym1, &symbol_mod);

		if (e1->token == T_UNEQUAL && e2->token == T_UNEQUAL &&
			   ((e1->right.sym == &symbol_yes && e2->right.sym == &symbol_mod) ||
			    (e1->right.sym == &symbol_mod && e2->right.sym == &symbol_yes)))
			// (a!='y') && (a!='m') -> (a='n')
			return expr_alloc_comp('=', sym1, &symbol_no);

		if (e1->token == T_UNEQUAL && e2->token == T_UNEQUAL &&
			   ((e1->right.sym == &symbol_mod && e2->right.sym == &symbol_no) ||
			    (e1->right.sym == &symbol_no && e2->right.sym == &symbol_mod)))
			// (a!='m') && (a!='n') -> (a='m')
			return expr_alloc_comp('=', sym1, &symbol_yes);

		if ((e1->token == T_WORD && e2->token == '=' && e2->right.sym == &symbol_mod) ||
		    (e2->token == T_WORD && e1->token == '=' && e1->right.sym == &symbol_mod) ||
		    (e1->token == T_WORD && e2->token == T_UNEQUAL && e2->right.sym == &symbol_yes) ||
		    (e2->token == T_WORD && e1->token == T_UNEQUAL && e1->right.sym == &symbol_yes))
			return NULL;
	}
	printf("optimize ");
	print_expr(0, e1, 0);
	printf(" && ");
	print_expr(0, e2, 0);
	printf(" ?\n");
	return NULL;
}

static void expr_eliminate_dups1(int token, struct expr **ep1, struct expr **ep2)
{
#define e1 (*ep1)
#define e2 (*ep2)
	struct expr *tmp;

	if (e1->token == token) {
		expr_eliminate_dups1(token, &e1->left.expr, &e2);
		expr_eliminate_dups1(token, &e1->right.expr, &e2);
		return;
	}
	if (e2->token == token) {
		expr_eliminate_dups1(token, &e1, &e2->left.expr);
		expr_eliminate_dups1(token, &e1, &e2->right.expr);
		return;
	}
	if (e1 == e2)
		return;

	switch (e1->token) {
	case '|': case '&':
		expr_eliminate_dups1(e1->token, &e1, &e1);
	}

	switch (token) {
	case '|':
		tmp = expr_join_or(e1, e2);
		if (tmp) {
			expr_free(e1); expr_free(e2);
			e1 = expr_alloc_symbol(&symbol_no);
			e2 = tmp;
			trans_count++;
		}
		break;
	case '&':
		tmp = expr_join_and(e1, e2);
		if (tmp) {
			expr_free(e1); expr_free(e2);
			e1 = expr_alloc_symbol(&symbol_yes);
			e2 = tmp;
			trans_count++;
		}
		break;
	}
#undef e1
#undef e2
}

static void expr_eliminate_dups2(int token, struct expr **ep1, struct expr **ep2)
{
#define e1 (*ep1)
#define e2 (*ep2)
	struct expr *tmp, *tmp1, *tmp2;

	if (e1->token == token) {
		expr_eliminate_dups2(token, &e1->left.expr, &e2);
		expr_eliminate_dups2(token, &e1->right.expr, &e2);
		return;
	}
	if (e2->token == token) {
		expr_eliminate_dups2(token, &e1, &e2->left.expr);
		expr_eliminate_dups2(token, &e1, &e2->right.expr);
	}
	if (e1 == e2)
		return;

	switch (e1->token) {
	case '|':
		expr_eliminate_dups2(e1->token, &e1, &e1);
		// (FOO || BAR) && (!FOO && !BAR) -> n
		tmp1 = expr_transform(expr_alloc_one('!', expr_copy(e1)));
		tmp2 = expr_copy(e2);
		tmp = expr_extract_eq_and(&tmp1, &tmp2);
		if (expr_is_yes(tmp1)) {
			expr_free(e1);
			e1 = expr_alloc_symbol(&symbol_no);
			trans_count++;
		}
		expr_free(tmp2);
		expr_free(tmp1);
		expr_free(tmp);
		break;
	case '&':
		expr_eliminate_dups2(e1->token, &e1, &e1);
		// (FOO && BAR) || (!FOO || !BAR) -> y
		tmp1 = expr_transform(expr_alloc_one('!', expr_copy(e1)));
		tmp2 = expr_copy(e2);
		tmp = expr_extract_eq_or(&tmp1, &tmp2);
		if (expr_is_no(tmp1)) {
			expr_free(e1);
			e1 = expr_alloc_symbol(&symbol_yes);
			trans_count++;
		}
		expr_free(tmp2);
		expr_free(tmp1);
		expr_free(tmp);
		break;
	}
#undef e1
#undef e2
}

struct expr *expr_eliminate_dups(struct expr *e)
{
	int oldcount;
	if (!e)
		return e;

	oldcount = trans_count;
	while (1) {
		trans_count = 0;
		switch (e->token) {
		case '|': case '&':
			expr_eliminate_dups1(e->token, &e, &e);
			expr_eliminate_dups2(e->token, &e, &e);
		}
		if (!trans_count)
			break;
		e = expr_eliminate_yn(e);
	}
	trans_count = oldcount;
	return e;
}

struct expr *expr_transform(struct expr *e)
{
	struct expr *tmp;

	if (!e)
		return NULL;
	switch (e->token) {
	case '=':
	case T_UNEQUAL:
	case T_WORD:
	case '^':
		break;
	default:
		e->left.expr = expr_transform(e->left.expr);
		e->right.expr = expr_transform(e->right.expr);
	}

	switch (e->token) {
	case '=':
		if (e->left.sym->type != SYMBOL_BOOLEAN)
			break;
		if (e->right.sym == &symbol_no) {
			e->token = '!';
			e->left.expr = expr_alloc_symbol(e->left.sym);
			e->right.sym = NULL;
			break;
		}
		if (e->right.sym == &symbol_mod) {
			printf("boolean symbol %s tested for 'm'? test forced to 'n'\n", e->left.sym->name);
			e->token = T_WORD;
			e->left.sym = &symbol_no;
			e->right.sym = NULL;
			break;
		}
		if (e->right.sym == &symbol_yes) {
			e->token = T_WORD;
			e->right.sym = NULL;
			break;
		}
		break;
	case T_UNEQUAL:
		if (e->left.sym->type != SYMBOL_BOOLEAN)
			break;
		if (e->right.sym == &symbol_no) {
			e->token = T_WORD;
			e->right.sym = NULL;
			break;
		}
		if (e->right.sym == &symbol_mod) {
			printf("boolean symbol %s tested for 'm'? test forced to 'y'\n", e->left.sym->name);
			e->token = T_WORD;
			e->left.sym = &symbol_yes;
			e->right.sym = NULL;
			break;
		}
		if (e->right.sym == &symbol_yes) {
			e->token = '!';
			e->left.expr = expr_alloc_symbol(e->left.sym);
			e->right.sym = NULL;
			break;
		}
		break;
	case '!':
		switch (e->left.expr->token) {
		case '!':
			// !!a -> a
			tmp = e->left.expr->left.expr;
			free(e->left.expr);
			free(e);
			e = tmp;
			e = expr_transform(e);
			break;
		case '=':
		case T_UNEQUAL:
			// !a='x' -> a!='x'
			tmp = e->left.expr;
			free(e);
			e = tmp;
			e->token = e->token == '=' ? T_UNEQUAL : '=';
			break;
		case '|':
			// !(a || b) -> !a && !b
			tmp = e->left.expr;
			e->token = '&';
			e->right.expr = expr_alloc_one('!', tmp->right.expr);
			tmp->token = '!';
			tmp->right.expr = NULL;
			e = expr_transform(e);
			break;
		case '&':
			// !(a && b) -> !a || !b
			tmp = e->left.expr;
			e->token = '|';
			e->right.expr = expr_alloc_one('!', tmp->right.expr);
			tmp->token = '!';
			tmp->right.expr = NULL;
			e = expr_transform(e);
			break;
		case T_WORD:
			if (e->left.expr->left.sym == &symbol_yes) {
				// !'y' -> 'n'
				tmp = e->left.expr;
				free(e);
				e = tmp;
				e->token = T_WORD;
				e->left.sym = &symbol_no;
				break;
			}
			if (e->left.expr->left.sym == &symbol_mod) {
				// !'m' -> 'm'
				tmp = e->left.expr;
				free(e);
				e = tmp;
				e->token = T_WORD;
				e->left.sym = &symbol_mod;
				break;
			}
			if (e->left.expr->left.sym == &symbol_no) {
				// !'n' -> 'y'
				tmp = e->left.expr;
				free(e);
				e = tmp;
				e->token = T_WORD;
				e->left.sym = &symbol_yes;
				break;
			}
			break;
		}
		break;
	}
	return e;
}

int expr_contains_symbol(struct expr *dep, struct symbol *sym)
{
	if (!dep)
		return 0;

	switch (dep->token) {
	case '&':
	case '|':
		return expr_contains_symbol(dep->left.expr, sym) ||
		       expr_contains_symbol(dep->right.expr, sym);
	case T_WORD:
		return dep->left.sym == sym;
	case '=':
	case T_UNEQUAL:
		return dep->left.sym == sym ||
		       dep->right.sym == sym;
	case '!':
		return expr_contains_symbol(dep->left.expr, sym);
	}
	return 0;
}

bool expr_depends_symbol(struct expr *dep, struct symbol *sym)
{
	if (!dep)
		return false;

	switch (dep->token) {
	case '&':
		return expr_depends_symbol(dep->left.expr, sym) ||
		       expr_depends_symbol(dep->right.expr, sym);
	case T_WORD:
		return dep->left.sym == sym;
	case '=':
		if (dep->left.sym == sym) {
			if (dep->right.sym == &symbol_yes || dep->right.sym == &symbol_mod)
				return true;
		}
		break;
	case T_UNEQUAL:
		if (dep->left.sym == sym) {
			if (dep->right.sym == &symbol_no)
				return true;
		}
		break;
	}
 	return false;
}

struct expr *expr_extract_eq_and(struct expr **ep1, struct expr **ep2)
{
	struct expr *tmp = NULL;
	expr_extract_eq('&', &tmp, ep1, ep2);
	if (tmp) {
		*ep1 = expr_eliminate_yn(*ep1);
		*ep2 = expr_eliminate_yn(*ep2);
	}
	return tmp;
}

struct expr *expr_extract_eq_or(struct expr **ep1, struct expr **ep2)
{
	struct expr *tmp = NULL;
	expr_extract_eq('|', &tmp, ep1, ep2);
	if (tmp) {
		*ep1 = expr_eliminate_yn(*ep1);
		*ep2 = expr_eliminate_yn(*ep2);
	}
	return tmp;
}

void expr_extract_eq(int token, struct expr **ep, struct expr **ep1, struct expr **ep2)
{
#define e1 (*ep1)
#define e2 (*ep2)
	if (e1->token == token) {
		expr_extract_eq(token, ep, &e1->left.expr, &e2);
		expr_extract_eq(token, ep, &e1->right.expr, &e2);
		return;
	}
	if (e2->token == token) {
		expr_extract_eq(token, ep, ep1, &e2->left.expr);
		expr_extract_eq(token, ep, ep1, &e2->right.expr);
		return;
	}
	if (expr_eq(e1, e2)) {
		*ep = *ep ? expr_alloc_two(token, *ep, e1) : e1;
		expr_free(e2);
		if (token == '&') {
			e1 = expr_alloc_symbol(&symbol_yes);
			e2 = expr_alloc_symbol(&symbol_yes);
		} else if (token == '|') {
			e1 = expr_alloc_symbol(&symbol_no);
			e2 = expr_alloc_symbol(&symbol_no);
		}
	}
#undef e1
#undef e2
}

struct expr *expr_trans_compare(struct expr *e, int token, struct symbol *sym)
{
	struct expr *e1, *e2;

	if (!e) {
		e = expr_alloc_symbol(sym);
		if (token == T_UNEQUAL)
			e = expr_alloc_one('!', e);
		return e;
	}
	switch (e->token) {
	case '&':
		e1 = expr_trans_compare(e->left.expr, '=', sym);
		e2 = expr_trans_compare(e->right.expr, '=', sym);
		if (sym == &symbol_yes)
			e = expr_alloc_two('&', e1, e2);
		if (sym == &symbol_no)
			e = expr_alloc_two('|', e1, e2);
		if (token == T_UNEQUAL)
			e = expr_alloc_one('!', e);
		return e;
	case '|':
		e1 = expr_trans_compare(e->left.expr, '=', sym);
		e2 = expr_trans_compare(e->right.expr, '=', sym);
		if (sym == &symbol_yes)
			e = expr_alloc_two('|', e1, e2);
		if (sym == &symbol_no)
			e = expr_alloc_two('&', e1, e2);
		if (token == T_UNEQUAL)
			e = expr_alloc_one('!', e);
		return e;
	case '!':
		return expr_trans_compare(e->left.expr, token == '=' ? T_UNEQUAL : '=', sym);
	case T_UNEQUAL:
	case '=':
		if (token == '=') {
			if (sym == &symbol_yes)
				return expr_copy(e);
			if (sym == &symbol_mod)
				return expr_alloc_symbol(&symbol_no);
			if (sym == &symbol_no)
				return expr_alloc_one('!', expr_copy(e));
		} else {
			if (sym == &symbol_yes)
				return expr_alloc_one('!', expr_copy(e));
			if (sym == &symbol_mod)
				return expr_alloc_symbol(&symbol_yes);
			if (sym == &symbol_no)
				return expr_copy(e);
		}
		break;
	case T_WORD:
		return expr_alloc_comp(token, e->left.sym, sym);
	}
	return NULL;
}

tristate expr_calc_value(struct expr *e)
{
	tristate val1, val2;
	const char *str1, *str2;

	if (!e)
		return yes;

	switch (e->token) {
	case T_WORD:
		sym_calc_value(e->left.sym);
		return S_TRI(e->left.sym->curr);
	case '&':
		val1 = expr_calc_value(e->left.expr);
		val2 = expr_calc_value(e->right.expr);
		return E_AND(val1, val2);
	case '|':
		val1 = expr_calc_value(e->left.expr);
		val2 = expr_calc_value(e->right.expr);
		return E_OR(val1, val2);
	case '!':
		val1 = expr_calc_value(e->left.expr);
		return E_NOT(val1);
	case '=':
		sym_calc_value(e->left.sym);
		sym_calc_value(e->right.sym);
		str1 = sym_get_string_value(e->left.sym);
		str2 = sym_get_string_value(e->right.sym);
		return !strcmp(str1, str2) ? yes : no;
	case T_UNEQUAL:
		sym_calc_value(e->left.sym);
		sym_calc_value(e->right.sym);
		str1 = sym_get_string_value(e->left.sym);
		str2 = sym_get_string_value(e->right.sym);
		return !strcmp(str1, str2) ? no : yes;
	default:
		printf("expr_calc_value: %d?\n", e->token);
		return no;
	}
}

static inline int token_gt(int t1, int t2)
{
#if 0
	return 1;
#else
	switch (t1) {
	case '=':
	case T_UNEQUAL:
		if (t2 == '!')
			return 1;
	case '!':
		if (t2 == '&')
			return 1;
	case '&':
		if (t2 == '|')
			return 1;
	case '|':
		if (t2 == '^')
			return 1;
	case '^':
		if (t2 == 0)
			return 1;
	case 0:
		return 0;
	}
	printf("[%dgt%d?]", t1, t2);
	return 0;
#endif
}

int print_type = 1;

static void print_sym(FILE *out, struct symbol *sym)
{
	fprintf(out, "%s", sym->name);
	if (print_type)
		fprintf(out, "%s",
			sym->type == SYMBOL_BOOLEAN ? "?" :
			sym->type == SYMBOL_TRISTATE ? "??" :
			sym->type == SYMBOL_HEX ? "h" :
			sym->type == SYMBOL_INT ? "i" :
			sym->type == SYMBOL_STRING ? "s" :
		       	"");
}

void fprint_expr(FILE *out, struct expr *e, int prevtoken)
{
	if (!e) {
		fprintf(out, "<none>");
		return;
	}

	switch (e->token) {
	case T_WORD:
		print_sym(out, e->left.sym);
		break;
	case '!':
		if (token_gt(prevtoken, '!'))
			fprintf(out, "(");
		fprintf(out, "!");
		fprint_expr(out, e->left.expr, '!');
		if (token_gt(prevtoken, '!'))
			fprintf(out, ")");
		break;
	case '=':
		if (token_gt(prevtoken, '='))
			fprintf(out, "(");
		print_sym(out, e->left.sym);
		fprintf(out, "=");
		print_sym(out, e->right.sym);
		if (token_gt(prevtoken, '='))
			fprintf(out, ")");
		break;
	case T_UNEQUAL:
		if (token_gt(prevtoken, T_UNEQUAL))
			fprintf(out, "(");
		print_sym(out, e->left.sym);
		fprintf(out, "!=");
		print_sym(out, e->right.sym);
		if (token_gt(prevtoken, T_UNEQUAL))
			fprintf(out, ")");
		break;
	case '|':
		if (token_gt(prevtoken, '|'))
			fprintf(out, "(");
		fprint_expr(out, e->left.expr, '|');
		fprintf(out, " || ");
		fprint_expr(out, e->right.expr, '|');
		if (token_gt(prevtoken, '|'))
			fprintf(out, ")");
		break;
	case '&':
		if (token_gt(prevtoken, '&'))
			fprintf(out, "(");
		fprint_expr(out, e->left.expr, '&');
		fprintf(out, " && ");
		fprint_expr(out, e->right.expr, '&');
		if (token_gt(prevtoken, '&'))
			fprintf(out, ")");
		break;
	case '^':
		if (e->left.expr) {
			fprint_expr(out, e->left.expr, '^');
			fprintf(out, " ^ ");
		}
		fprintf(out, "%s", e->right.sym->name);
		break;
	default:
		fprintf(out, "<unknown token %d>", e->token);
		break;
	}
}

void print_expr(int mask, struct expr *e, int prevtoken)
{
	if (!(cdebug & mask))
		return;
	fprint_expr(stdout, e, prevtoken);
}

